/*============================================================================
             DEPTH FIRST VISIT OF A TREE (recursive version)
Grammar for the input tree:  char ::= '0' | '1' | ... | '9'
                             tree ::= char | - char tree | (tree char tree)
Internal tree:               tree = Num  |  Num x tree  |  tree x Num x tree
                                   (Leaf)     (Unary)          (Binary)
Grammar for the output tree: tree ::= n | (n.tree) | (tree.n.tree)
The program uses a stack of trees, implicit in the recursion.
Thanks to Nicholas Amadori.
=========================================================================== */
#include <iostream.h>
#include <stdlib.h>   // file needed for the EXIT_SUCCESS and the EXIT_FAILURE
//----------------------------------------------------------------------------
//                    trees
typedef int Values;
class tree {
public :
virtual void dfvisit()=0;    // dfvisit is depth_first_visit
virtual void print()=0;      // print function
};
//---------------------------
class Leaf: public tree{
public: Values value;
        void dfvisit(); void print();};
void Leaf::dfvisit() {cout << value << " ";};
void Leaf::print()   {cout << value;};
//---------------------------
class Unary: public tree{
public: Values value1; tree*  son;
        void dfvisit(); void print();};
void Unary::dfvisit() {(*son).dfvisit(); cout << value1 << " ";};
void Unary::print()   {cout << "(" << value1 << ".";
                       (*son).print(); cout << ")";};
//---------------------------
class Binary: public tree{
public: Values value2; tree* left; tree* right;
        void dfvisit(); void print();};
void Binary::dfvisit() {(*left).dfvisit(); (*right).dfvisit();
                        cout << value2 << " ";};
void Binary::print() {cout << "("; (*left).print();
                      cout << "." << value2 << ".";
                      (*right).print(); cout << ")";};
//----------------------------------------------------------------------------
//                    from characters to integers
int v(char c){return c-48;};
//----------------------------------------------------------------------------
//                    parsing trees
tree* parsetree()
{char c; cin >> c;
if ('0' <= c && c <= '9') {Leaf* l=new Leaf; (*l).value=v(c); return l;} else
if (c == '-') {Unary* u=new Unary; cin >> c; (*u).value1=v(c);
               (*u).son=parsetree(); return u;} else
if (c == '(') {Binary* b=new Binary; (*b).left=parsetree(); cin >> c;
               (*b).value2=v(c); (*b).right=parsetree(); cin >> c; return b;}
else {cerr << "Error in parsetree!"; exit(EXIT_FAILURE);}
};
//----------------------------------------------------------------------------
int main()
{tree* t;
cout << "Input a tree according to the given grammar above, please.\n";
t=parsetree(); cout << "\nTree in input:\n"; (*t).print();
cout << "\n\nDepth first visit of the tree\n";
(*t).dfvisit(); cout << endl; return(EXIT_SUCCESS);
}
/* input:    output:
------------------------------------------------------------------------------
             Input a tree according to the given grammar above, please.
             (-231(-564-78))
 
             Tree in input:
             ((2.3).1.((5.6).4.(7.8)))
 
             Depth first visit of the tree:
             3 2 6 5 8 7 4 1     
--------------------------------------------------------------------------- */